/*
 * Open file cache.
 *
 * (c) 2015 - Jeff Layton <jeff.layton@primarydata.com>
 */

#include <linux/hash.h>
#include <linux/slab.h>
#include <linux/hash.h>
#include <linux/file.h>
#include <linux/sched.h>
#include <linux/list_lru.h>
#include <linux/fsnotify_backend.h>
#include <linux/seq_file.h>

#include "vfs.h"
#include "nfsd.h"
#include "nfsfh.h"
#include "filecache.h"
#include "trace.h"

#define NFSDDBG_FACILITY	NFSDDBG_FH

/* FIXME: dynamically size this for the machine somehow? */
#define NFSD_FILE_HASH_BITS                   12
#define NFSD_FILE_HASH_SIZE                  (1 << NFSD_FILE_HASH_BITS)

/* We only care about NFSD_MAY_READ/WRITE for this cache */
#define NFSD_FILE_MAY_MASK	(NFSD_MAY_READ|NFSD_MAY_WRITE)

struct nfsd_fcache_bucket {
	struct hlist_head	nfb_head;
	spinlock_t		nfb_lock;
	unsigned int		nfb_count;
	unsigned int		nfb_maxcount;
};

static struct kmem_cache		*nfsd_file_slab;
static struct kmem_cache		*nfsd_file_mark_slab;
static struct nfsd_fcache_bucket	*nfsd_file_hashtbl;
static struct list_lru			nfsd_file_lru;
static struct fsnotify_group		*nfsd_file_fsnotify_group;

static void
nfsd_file_slab_free(struct rcu_head *rcu)
{
	struct nfsd_file *nf = container_of(rcu, struct nfsd_file, nf_rcu);

	kmem_cache_free(nfsd_file_slab, nf);
}

static void
nfsd_file_mark_free(struct fsnotify_mark *mark)
{
	struct nfsd_file_mark *nfm = container_of(mark, struct nfsd_file_mark,
						  nfm_mark);

	kmem_cache_free(nfsd_file_mark_slab, nfm);
}

static struct nfsd_file_mark *
nfsd_file_mark_get(struct nfsd_file_mark *nfm)
{
	if (!atomic_inc_not_zero(&nfm->nfm_ref))
		return NULL;
	return nfm;
}

static void
nfsd_file_mark_put(struct nfsd_file_mark *nfm)
{
	if (atomic_dec_and_test(&nfm->nfm_ref))
		fsnotify_destroy_mark(&nfm->nfm_mark, nfsd_file_fsnotify_group);
}

static struct nfsd_file_mark *
nfsd_file_mark_find_or_create(struct nfsd_file *nf, struct inode *inode)
{
	int			err;
	struct fsnotify_mark	*mark;
	struct nfsd_file_mark	*nfm = NULL, *new = NULL;

	do {
		mark = fsnotify_find_inode_mark(nfsd_file_fsnotify_group,
						inode);
		if (mark) {
			nfm = nfsd_file_mark_get(container_of(mark,
						 struct nfsd_file_mark,
						 nfm_mark));
			fsnotify_put_mark(mark);
			if (likely(nfm))
				break;
		}

		/* allocate a new nfm */
		if (!new) {
			new = kmem_cache_alloc(nfsd_file_mark_slab, GFP_KERNEL);
			if (!new)
				return NULL;
			fsnotify_init_mark(&new->nfm_mark, nfsd_file_mark_free);
			atomic_set(&new->nfm_ref, 1);
		}

		err = fsnotify_add_mark(&new->nfm_mark,
				nfsd_file_fsnotify_group, nf->nf_inode,
				NULL, false);
		if (likely(!err)) {
			nfm = new;
			new = NULL;
		}
	} while (unlikely(err == EEXIST));

	if (new)
		kmem_cache_free(nfsd_file_mark_slab, new);
	return nfm;
}

static struct nfsd_file *
nfsd_file_alloc(struct inode *inode, unsigned int may, unsigned int hashval)
{
	struct nfsd_file *nf;

	nf = kmem_cache_alloc(nfsd_file_slab, GFP_KERNEL);
	if (nf) {
		INIT_HLIST_NODE(&nf->nf_node);
		INIT_LIST_HEAD(&nf->nf_lru);
		nf->nf_file = NULL;
		nf->nf_flags = 0;
		nf->nf_inode = inode;
		nf->nf_hashval = hashval;
		atomic_set(&nf->nf_ref, 1);
		nf->nf_may = NFSD_FILE_MAY_MASK & may;
		if (may & NFSD_MAY_NOT_BREAK_LEASE) {
			if (may & NFSD_MAY_WRITE)
				__set_bit(NFSD_FILE_BREAK_WRITE, &nf->nf_flags);
			if (may & NFSD_MAY_READ)
				__set_bit(NFSD_FILE_BREAK_READ, &nf->nf_flags);
		}
		nf->nf_mark = NULL;
		trace_nfsd_file_alloc(nf);
	}
	return nf;
}

static void
nfsd_file_put_final(struct nfsd_file *nf)
{
	trace_nfsd_file_put_final(nf);
	if (nf->nf_mark)
		nfsd_file_mark_put(nf->nf_mark);
	if (nf->nf_file)
		fput(nf->nf_file);
	call_rcu(&nf->nf_rcu, nfsd_file_slab_free);
}

static bool
nfsd_file_put_final_delayed(struct nfsd_file *nf)
{
	bool flush = false;

	trace_nfsd_file_put_final(nf);
	if (nf->nf_mark)
		nfsd_file_mark_put(nf->nf_mark);
	if (nf->nf_file)
		flush = fput_global(nf->nf_file);
	call_rcu(&nf->nf_rcu, nfsd_file_slab_free);
	return flush;
}

static bool
nfsd_file_unhash(struct nfsd_file *nf)
{
	lockdep_assert_held(&nfsd_file_hashtbl[nf->nf_hashval].nfb_lock);

	trace_nfsd_file_unhash(nf);
	if (test_bit(NFSD_FILE_HASHED, &nf->nf_flags)) {
		--nfsd_file_hashtbl[nf->nf_hashval].nfb_count;
		clear_bit(NFSD_FILE_HASHED, &nf->nf_flags);
		hlist_del_rcu(&nf->nf_node);
		list_lru_del(&nfsd_file_lru, &nf->nf_lru);
		return true;
	}
	return false;
}

static void
nfsd_file_unhash_and_release_locked(struct nfsd_file *nf, struct list_head *dispose)
{
	lockdep_assert_held(&nfsd_file_hashtbl[nf->nf_hashval].nfb_lock);

	trace_nfsd_file_unhash_and_release_locked(nf);
	if (!nfsd_file_unhash(nf))
		return;
	if (!atomic_dec_and_test(&nf->nf_ref))
		return;

	list_add(&nf->nf_lru, dispose);
}

void
nfsd_file_put(struct nfsd_file *nf)
{
	trace_nfsd_file_put(nf);
	set_bit(NFSD_FILE_REFERENCED, &nf->nf_flags);
	smp_mb__after_atomic();
	if (atomic_dec_and_test(&nf->nf_ref)) {
		WARN_ON(test_bit(NFSD_FILE_HASHED, &nf->nf_flags));
		nfsd_file_put_final(nf);
	}
}

struct nfsd_file *
nfsd_file_get(struct nfsd_file *nf)
{
	if (likely(atomic_inc_not_zero(&nf->nf_ref)))
		return nf;
	return NULL;
}

static void
nfsd_file_dispose_list(struct list_head *dispose)
{
	struct nfsd_file *nf;

	while(!list_empty(dispose)) {
		nf = list_first_entry(dispose, struct nfsd_file, nf_lru);
		list_del(&nf->nf_lru);
		nfsd_file_put_final(nf);
	}
}

static void
nfsd_file_dispose_list_sync(struct list_head *dispose)
{
	bool flush = false;
	struct nfsd_file *nf;

	while(!list_empty(dispose)) {
		nf = list_first_entry(dispose, struct nfsd_file, nf_lru);
		list_del(&nf->nf_lru);
		if (nfsd_file_put_final_delayed(nf))
			flush = true;
	}
	if (flush)
		fput_global_flush();
}

static enum lru_status
nfsd_file_lru_cb(struct list_head *item, struct list_lru_one *lru,
		 spinlock_t *lock, void *arg)
	__releases(lock)
	__acquires(lock)
{
	struct nfsd_file *nf = list_entry(item, struct nfsd_file, nf_lru);
	bool unhashed;

	/*
	 * Do a lockless refcount check. The hashtable holds one reference, so
	 * we look to see if anything else has a reference, or if any have
	 * been put since the shrinker last ran. Those don't get unhashed and
	 * released.
	 *
	 * Note that in the put path, we set the flag and then decrement the
	 * counter. Here we check the counter and then test and clear the flag.
	 * That order is deliberate to ensure that we can do this locklessly.
	 */
	if (atomic_read(&nf->nf_ref) > 1 ||
	    test_and_clear_bit(NFSD_FILE_REFERENCED, &nf->nf_flags))
		return LRU_ROTATE;

	spin_unlock(lock);
	spin_lock(&nfsd_file_hashtbl[nf->nf_hashval].nfb_lock);
	unhashed = nfsd_file_unhash(nf);
	spin_unlock(&nfsd_file_hashtbl[nf->nf_hashval].nfb_lock);
	if (unhashed)
		nfsd_file_put(nf);
	spin_lock(lock);
	return unhashed ? LRU_REMOVED_RETRY : LRU_RETRY;
}

static unsigned long
nfsd_file_lru_count(struct shrinker *s, struct shrink_control *sc)
{
	return list_lru_count(&nfsd_file_lru);
}

static unsigned long
nfsd_file_lru_scan(struct shrinker *s, struct shrink_control *sc)
{
	return list_lru_shrink_walk(&nfsd_file_lru, sc, nfsd_file_lru_cb, NULL);
}

static struct shrinker	nfsd_file_shrinker = {
	.scan_objects = nfsd_file_lru_scan,
	.count_objects = nfsd_file_lru_count,
	.seeks = 1,
};

static void
__nfsd_file_close_inode(struct inode *inode, unsigned int hashval,
			struct list_head *dispose)
{
	struct nfsd_file	*nf;
	struct hlist_node	*tmp;

	spin_lock(&nfsd_file_hashtbl[hashval].nfb_lock);
	hlist_for_each_entry_safe(nf, tmp, &nfsd_file_hashtbl[hashval].nfb_head, nf_node) {
		if (inode == nf->nf_inode)
			nfsd_file_unhash_and_release_locked(nf, dispose);
	}
	spin_unlock(&nfsd_file_hashtbl[hashval].nfb_lock);
}

/**
 * nfsd_file_close_inode_sync - attempt to forcibly close a nfsd_file
 * @inode: inode of the file to attempt to remove
 *
 * Walk the whole hash bucket, looking for any files that correspond to "inode".
 * If any do, then unhash them and put the hashtable reference to them and
 * destroy any that had their last reference put. Also ensure that any of the
 * fputs also have their final __fput done as well.
 */
void
nfsd_file_close_inode_sync(struct inode *inode)
{
	unsigned int		hashval = (unsigned int)hash_long(inode->i_ino,
						NFSD_FILE_HASH_BITS);
	LIST_HEAD(dispose);

	__nfsd_file_close_inode(inode, hashval, &dispose);
	trace_nfsd_file_close_inode_sync(inode, hashval, !list_empty(&dispose));
	nfsd_file_dispose_list_sync(&dispose);
}

/**
 * nfsd_file_close_inode_sync - attempt to forcibly close a nfsd_file
 * @inode: inode of the file to attempt to remove
 *
 * Walk the whole hash bucket, looking for any files that correspond to "inode".
 * If any do, then unhash them and put the hashtable reference to them and
 * destroy any that had their last reference put.
 */
static void
nfsd_file_close_inode(struct inode *inode)
{
	unsigned int		hashval = (unsigned int)hash_long(inode->i_ino,
						NFSD_FILE_HASH_BITS);
	LIST_HEAD(dispose);

	__nfsd_file_close_inode(inode, hashval, &dispose);
	trace_nfsd_file_close_inode(inode, hashval, !list_empty(&dispose));
	nfsd_file_dispose_list(&dispose);
}

static int
nfsd_file_lease_notifier_call(struct notifier_block *nb, unsigned long arg,
			    void *data)
{
	struct file_lock *fl = data;

	/* Only close files for F_SETLEASE leases */
	if (fl->fl_flags & FL_LEASE)
		nfsd_file_close_inode_sync(file_inode(fl->fl_file));
	return 0;
}

static struct notifier_block nfsd_file_lease_notifier = {
	.notifier_call = nfsd_file_lease_notifier_call,
};

static int
nfsd_file_fsnotify_handle_event(struct fsnotify_group *group,
				struct inode *inode,
				struct fsnotify_mark *inode_mark,
				struct fsnotify_mark *vfsmount_mark,
				u32 mask, void *data, int data_type,
				const unsigned char *file_name, u32 cookie)
{
	trace_nfsd_file_fsnotify_handle_event(inode, mask);

	/* Should be no marks on non-regular files */
	if (!S_ISREG(inode->i_mode)) {
		WARN_ON_ONCE(1);
		return 0;
	}

	/* ...and we don't do anything with vfsmount marks */
	BUG_ON(vfsmount_mark);

	/* don't close files if this was not the last link */
	if (mask & FS_ATTRIB) {
		if (inode->i_nlink)
			return 0;
	}

	nfsd_file_close_inode(inode);
	return 0;
}


const static struct fsnotify_ops nfsd_file_fsnotify_ops = {
	.handle_event = nfsd_file_fsnotify_handle_event,
};

int
nfsd_file_cache_init(void)
{
	int		ret = -ENOMEM;
	unsigned int	i;

	if (nfsd_file_hashtbl)
		return 0;

	nfsd_file_hashtbl = kcalloc(NFSD_FILE_HASH_SIZE,
				sizeof(*nfsd_file_hashtbl), GFP_KERNEL);
	if (!nfsd_file_hashtbl) {
		pr_err("nfsd: unable to allocate nfsd_file_hashtbl\n");
		goto out_err;
	}

	nfsd_file_slab = kmem_cache_create("nfsd_file",
				sizeof(struct nfsd_file), 0, 0, NULL);
	if (!nfsd_file_slab) {
		pr_err("nfsd: unable to create nfsd_file_slab\n");
		goto out_err;
	}

	nfsd_file_mark_slab = kmem_cache_create("nfsd_file_mark",
					sizeof(struct nfsd_file_mark), 0, 0, NULL);
	if (!nfsd_file_mark_slab) {
		pr_err("nfsd: unable to create nfsd_file_mark_slab\n");
		goto out_err;
	}


	ret = list_lru_init(&nfsd_file_lru);
	if (ret) {
		pr_err("nfsd: failed to init nfsd_file_lru: %d\n", ret);
		goto out_err;
	}

	ret = register_shrinker(&nfsd_file_shrinker);
	if (ret) {
		pr_err("nfsd: failed to register nfsd_file_shrinker: %d\n", ret);
		goto out_lru;
	}

	ret = srcu_notifier_chain_register(&lease_notifier_chain,
				&nfsd_file_lease_notifier);
	if (ret) {
		pr_err("nfsd: unable to register lease notifier: %d\n", ret);
		goto out_shrinker;
	}

	nfsd_file_fsnotify_group = fsnotify_alloc_group(&nfsd_file_fsnotify_ops);
	if (IS_ERR(nfsd_file_fsnotify_group)) {
		pr_err("nfsd: unable to create fsnotify group: %ld\n",
			PTR_ERR(nfsd_file_fsnotify_group));
		nfsd_file_fsnotify_group = NULL;
		goto out_notifier;
	}

	for (i = 0; i < NFSD_FILE_HASH_SIZE; i++) {
		INIT_HLIST_HEAD(&nfsd_file_hashtbl[i].nfb_head);
		spin_lock_init(&nfsd_file_hashtbl[i].nfb_lock);
	}
out:
	return ret;
out_notifier:
	srcu_notifier_chain_unregister(&lease_notifier_chain,
				&nfsd_file_lease_notifier);
out_shrinker:
	unregister_shrinker(&nfsd_file_shrinker);
out_lru:
	list_lru_destroy(&nfsd_file_lru);
out_err:
	kmem_cache_destroy(nfsd_file_slab);
	nfsd_file_slab = NULL;
	kmem_cache_destroy(nfsd_file_mark_slab);
	nfsd_file_mark_slab = NULL;
	kfree(nfsd_file_hashtbl);
	nfsd_file_hashtbl = NULL;
	goto out;
}

void
nfsd_file_cache_purge(void)
{
	unsigned int		i;
	struct nfsd_file	*nf;
	LIST_HEAD(dispose);

	if (!nfsd_file_hashtbl)
		return;

	for (i = 0; i < NFSD_FILE_HASH_SIZE; i++) {
		spin_lock(&nfsd_file_hashtbl[i].nfb_lock);
		while(!hlist_empty(&nfsd_file_hashtbl[i].nfb_head)) {
			nf = hlist_entry(nfsd_file_hashtbl[i].nfb_head.first,
					 struct nfsd_file, nf_node);
			nfsd_file_unhash_and_release_locked(nf, &dispose);
		}
		spin_unlock(&nfsd_file_hashtbl[i].nfb_lock);
		nfsd_file_dispose_list(&dispose);
	}
}

void
nfsd_file_cache_shutdown(void)
{
	LIST_HEAD(dispose);

	srcu_notifier_chain_unregister(&lease_notifier_chain,
				&nfsd_file_lease_notifier);
	unregister_shrinker(&nfsd_file_shrinker);
	nfsd_file_cache_purge();
	list_lru_destroy(&nfsd_file_lru);
	rcu_barrier();
	fsnotify_put_group(nfsd_file_fsnotify_group);
	nfsd_file_fsnotify_group = NULL;
	kmem_cache_destroy(nfsd_file_slab);
	nfsd_file_slab = NULL;
	kmem_cache_destroy(nfsd_file_mark_slab);
	nfsd_file_mark_slab = NULL;
	kfree(nfsd_file_hashtbl);
	nfsd_file_hashtbl = NULL;
}

/*
 * Search nfsd_file_hashtbl[] for file. We hash on the filehandle and also on
 * the NFSD_MAY_READ/WRITE flags. If the file is open for r/w, then it's usable
 * for either.
 */
static struct nfsd_file *
nfsd_file_find_locked(struct inode *inode, unsigned int may_flags,
			unsigned int hashval)
{
	struct nfsd_file *nf;
	unsigned char need = may_flags & NFSD_FILE_MAY_MASK;

	hlist_for_each_entry_rcu(nf, &nfsd_file_hashtbl[hashval].nfb_head,
				 nf_node) {
		if ((need & nf->nf_may) != need)
			continue;
		if (nf->nf_inode == inode)
			return nfsd_file_get(nf);
	}
	return NULL;
}

__be32
nfsd_file_acquire(struct svc_rqst *rqstp, struct svc_fh *fhp,
		  unsigned int may_flags, struct nfsd_file **pnf)
{
	__be32	status;
	struct nfsd_file *nf, *new = NULL;
	struct inode *inode;
	unsigned int hashval;

	/* FIXME: skip this if fh_dentry is already set? */
	status = fh_verify(rqstp, fhp, S_IFREG,
				may_flags|NFSD_MAY_OWNER_OVERRIDE);
	if (status != nfs_ok)
		return status;

	inode = d_inode(fhp->fh_dentry);
	hashval = (unsigned int)hash_long(inode->i_ino, NFSD_FILE_HASH_BITS);
retry:
	rcu_read_lock();
	nf = nfsd_file_find_locked(inode, may_flags, hashval);
	rcu_read_unlock();
	if (nf)
		goto wait_for_construction;

	if (!new) {
		new = nfsd_file_alloc(inode, may_flags, hashval);
		if (!new) {
			trace_nfsd_file_acquire(hashval, inode, may_flags, NULL,
						nfserr_jukebox);
			return nfserr_jukebox;
		}
	}

	spin_lock(&nfsd_file_hashtbl[hashval].nfb_lock);
	nf = nfsd_file_find_locked(inode, may_flags, hashval);
	if (likely(nf == NULL)) {
		/* Take reference for the hashtable */
		atomic_inc(&new->nf_ref);
		__set_bit(NFSD_FILE_HASHED, &new->nf_flags);
		__set_bit(NFSD_FILE_PENDING, &new->nf_flags);
		list_lru_add(&nfsd_file_lru, &new->nf_lru);
		hlist_add_head_rcu(&new->nf_node,
				&nfsd_file_hashtbl[hashval].nfb_head);
		++nfsd_file_hashtbl[hashval].nfb_count;
		nfsd_file_hashtbl[hashval].nfb_maxcount = max(nfsd_file_hashtbl[hashval].nfb_maxcount,
				nfsd_file_hashtbl[hashval].nfb_count);
		spin_unlock(&nfsd_file_hashtbl[hashval].nfb_lock);
		nf = new;
		new = NULL;
		goto open_file;
	}
	spin_unlock(&nfsd_file_hashtbl[hashval].nfb_lock);

wait_for_construction:
	wait_on_bit(&nf->nf_flags, NFSD_FILE_PENDING, TASK_UNINTERRUPTIBLE);

	/* Did construction of this file fail? */
	if (!nf->nf_file) {
		/*
		 * We can only take over construction for this nfsd_file if the
		 * MAY flags are equal. Otherwise, we put the reference and try
		 * again.
		 */
		if ((may_flags & NFSD_FILE_MAY_MASK) != nf->nf_may) {
			nfsd_file_put(nf);
			goto retry;
		}

		/* try to take over construction for this file */
		if (test_and_set_bit(NFSD_FILE_PENDING, &nf->nf_flags))
			goto wait_for_construction;

		/* sync up the BREAK_* flags with our may_flags */
		if (may_flags & NFSD_MAY_NOT_BREAK_LEASE) {
			if (may_flags & NFSD_MAY_WRITE)
				set_bit(NFSD_FILE_BREAK_WRITE, &nf->nf_flags);
			if (may_flags & NFSD_MAY_READ)
				set_bit(NFSD_FILE_BREAK_READ, &nf->nf_flags);
		} else {
			clear_bit(NFSD_FILE_BREAK_WRITE, &nf->nf_flags);
			clear_bit(NFSD_FILE_BREAK_READ, &nf->nf_flags);
		}

		goto open_file;
	}

	if (!(may_flags & NFSD_MAY_NOT_BREAK_LEASE)) {
		bool write = (may_flags & NFSD_MAY_WRITE);

		if (test_bit(NFSD_FILE_BREAK_READ, &nf->nf_flags) ||
		    (test_bit(NFSD_FILE_BREAK_WRITE, &nf->nf_flags) && write)) {
			status = nfserrno(nfsd_open_break_lease(
					file_inode(nf->nf_file), may_flags));
			if (status == nfs_ok) {
				clear_bit(NFSD_FILE_BREAK_READ, &nf->nf_flags);
				if (write)
					clear_bit(NFSD_FILE_BREAK_WRITE,
						  &nf->nf_flags);
			}
		}
	}
out:
	if (status == nfs_ok) {
		*pnf = nf;
	} else {
		nfsd_file_put(nf);
		nf = NULL;
	}

	if (new)
		nfsd_file_put(new);

	trace_nfsd_file_acquire(hashval, inode, may_flags, nf, status);
	return status;
open_file:
	if (!nf->nf_mark) {
		nf->nf_mark = nfsd_file_mark_find_or_create(nf, inode);
		if (!nf->nf_mark)
			status = nfserr_jukebox;
	}
	/* FIXME: should we abort opening if the link count goes to 0? */
	if (status == nfs_ok)
		status = nfsd_open_verified(rqstp, fhp, S_IFREG, may_flags,
						&nf->nf_file);
	clear_bit_unlock(NFSD_FILE_PENDING, &nf->nf_flags);
	smp_mb__after_atomic();
	wake_up_bit(&nf->nf_flags, NFSD_FILE_PENDING);
	goto out;
}

/*
 * Note that fields may be added, removed or reordered in the future. Programs
 * scraping this file for info should test the labels to ensure they're
 * getting the correct field.
 */
static int nfsd_file_cache_stats_show(struct seq_file *m, void *v)
{
	unsigned int i, count = 0, longest = 0;

	/*
	 * No need for spinlocks here since we're not terribly interested in
	 * accuracy. We do take the nfsd_mutex simply to ensure that we
	 * don't end up racing with server shutdown
	 */
	mutex_lock(&nfsd_mutex);
	if (nfsd_file_hashtbl) {
		for (i = 0; i < NFSD_FILE_HASH_SIZE; i++) {
			count += nfsd_file_hashtbl[i].nfb_count;
			longest = max(longest, nfsd_file_hashtbl[i].nfb_count);
		}
	}
	mutex_unlock(&nfsd_mutex);

	seq_printf(m, "total entries: %u\n", count);
	seq_printf(m, "longest chain: %u\n", longest);
	return 0;
}

int nfsd_file_cache_stats_open(struct inode *inode, struct file *file)
{
	return single_open(file, nfsd_file_cache_stats_show, NULL);
}
